"""
Tools for model handling
"""

from sklearn.metrics import mean_squared_error
import numpy as np


def get_operation_model(model, X, y, metric=mean_squared_error, **kargv):
    """
    def get_operation_model(model, X, y, metric=mean_squared_error, *argv)

    Selects an operation model by running the entire dataset for each fold
    and picking the one with the best metric.

    The function always assume that the best value for a metric is its minimum value
    (MSE, for instance). If you use a metric like R2, where the higher the better, you must
    multiply such metric by -1.

    Input parameters:
        - model: a model structure as returned by sklearn cross_validation.
        - X: the input dataset
        - y: the target values
        - metric: a metric to be used (MSE, R2, etc.)
        - **kargv: passed down to the metric function.
    """
    res = np.array([metric(y, mod.predict(X), **kargv) for mod in model['estimator']])
    op_idx = np.argmin(res)
    return model['estimator'][op_idx]


def create_validation_dataset(model, X, y, proba: bool=True, copy: bool=True):
    """
    def create_validation_dataset(model, X, y, proba: bool=True, copy: bool=True)

    Create a validation dataset suitable for validate your model
    when deployed to another operation environment.

    Input parameters are:
        - model: a sklearn-like model (pipeline, skorch, etc.)
        - X: the input dataset
        - y: the targets of X
        - proba: if True, will call model.predict_proba. Otherwise, model.predict
        - copy: whether to return a copy of the passed dataset (both X and y).
    
    Returns:
        A new dataset containing 2 new columns:
            - y_true: containing the targets in y.
            - y_pred: the model output for each sample in X.
    """

    if copy: X = X.copy()
    X['y_true'] = y
    X['y_pred'] = model.predict_proba(X) if proba else model.predict(X)
    return X


def get_input_variables_description(X):
    """
    Create a descriptive pandas telling the following info from X:
      - The order of each column (Inpout Order)
      - Column names (Input Name)
      - Column types (Input Type)
    """
    input_description = X.dtypes.to_frame().reset_index()
    input_description.index.name = 'Input Order'
    input_description.columns = ['Input Name', 'Input Type']
    return input_description




"""Metric related tools"""

import numpy as np
from scipy.stats import gmean
from seaborn.algorithms import bootstrap
from sklearn.metrics import recall_score

def bootstrap_estimate(vec, ci=95, n_boot=1000, seed=None):
    """
    def bootstrap_estimate(vec, ci=95, n_boot=1000)

    Returns the aggregated result for vector vec using the same CI estimator as seaborn.

    Input:
      - vec: a numpy vector [N,]
      - ci: the confidence interval to consider.
      - n_boot: how many samplings to employ when using bootstrap for the CI interval.
      - seed: the seed value to use.

    Returns: a tuple with the following values:
      - The mean value of vec
      - The lower limit of the confidence interval
      - The upper limit of the confidence interval
    """
    def percentile_interval(data, width):
        """Return a percentile interval from data of a given width."""
        edge = (100 - width) / 2
        percentiles = edge, 100 - edge
        return np.percentile(data, percentiles)

    mean = vec.mean()
    boots = bootstrap(vec, func='mean', n_boot=n_boot, seed=seed)
    err_min, err_max = percentile_interval(boots, ci)

    return mean, err_min, err_max


def sp_index(tp: np.array, tn: np.array) -> np.array:
  """
  def sp(tp: np.array, tn: np.array) -> np.array

  Calculates the SP index, which is given by:

  sp = \sqrt{ \sqrt{tp \times tn} \times \(\frac{tp+tn,2}\) }

  where tp is the true positive values and tn the true negative values.

  Returns: an array with the sp index calculated.
  """

  if type(tp) is np.ndarray: tp = tp.flatten()
  if type(tn) is np.ndarray: tn = tn.flatten()
  mat = np.array([tp, tn])
  return np.sqrt( gmean(mat, axis=0) * mat.mean(axis=0) ).flatten()


def sensitivity(y_true, y_pred):
  """
  def sensitivity(y_true, y_pred)

  Calculate the sensitivity score. Sklearn style.
  """
  return recall_score(y_true, y_pred, pos_label=1)


def specificity(y_true, y_pred):
  """
  def specificity(y_true, y_pred)

  Calculate the specificity score. Sklearn style.
  """
  return recall_score(y_true, y_pred, pos_label=0)


def sp_score(y_true, y_pred):
  """
  def sp_score(y_true, y_pred)

  Calculate the sp_score score. Sklearn style.
  """
  return sp_index(sensitivity(y_true, y_pred), specificity(y_true, y_pred))
